from datetime import timedelta

from django.apps import apps
from django.db import transaction

from whisper import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import QuerySet, Count, Q, F
from django.template.defaultfilters import title
from django.utils.timezone import now


class RoomQuerySet(QuerySet):
    def create_from_slug(self, slug, creator=None):
        room_name, users = self.get_room_name_and_users_from_slug(slug)
        room = self.update_or_create(slug=slug, defaults={'name': room_name})[0]

        from whisper.models import RoomUser
        for user in users:
            RoomUser.objects.get_or_create(room=room, user=user, defaults={'last_read': now()})

        if creator:
            RoomUser.objects.get_or_create(room=room, user=creator, defaults={'last_read': now()})

        return room

    @transaction.atomic
    def create_group(self):
        room = self.create()
        room.slug = "group-" + str(room.pk)
        room.name = "Group #" + str(room.pk)
        room.save(update_fields=['slug', 'name'])
        return room

    def get_or_create_from_users(self, user_ids):
        users_room = self.filter(slug__startswith='group').annotate(cnt=Count('users')).filter(cnt=len(user_ids))

        for user_id in user_ids:
            users_room = users_room.filter(users__pk=user_id)

        if users_room:
            return users_room.latest()
        else:
            new_room = self.create_group()

            from whisper.models import RoomUser
            for user_id in user_ids:
                RoomUser.objects.get_or_create(room=new_room, user_id=user_id, defaults={'last_read': now()})

            return new_room

    @staticmethod
    def get_room_name_and_users_from_slug(slug):
        slug_parts = slug.split('-', 1)
        subject = slug_parts[0]
        users = get_user_model().objects.none()

        if subject == 'users':
            user_pks = slug_parts[1].split('-')
            users = get_user_model().objects.filter(pk__in=user_pks)
            room_name = ' & '.join(list(map(str, users)))
        else:
            subject_pk = slug_parts[1]
            obj_name = f'#{subject_pk}'

            for model in apps.get_models():
                if subject == model.__name__.lower():
                    model_name = model._meta.verbose_name
                    obj = model.objects.get(pk=subject_pk)
                    users = obj.room_members
                    obj_name = str(obj)

            room_name = title(f'{model_name} {obj_name}')

        return room_name, users

    def rename_by_slug(self, slug, new_name, default_name=None):
        try:
            room = self.get(slug=slug)

            if room.name == default_name or default_name is None:
                room.name = new_name
                room.save(update_fields=['name'])

        except ObjectDoesNotExist:
            pass

    def of_user(self, user):
        if not user.is_authenticated:
            return self.none()

        return user.room_set.all()

    def count_unread_messages(self, user):
        if not user.is_authenticated:
            return self.none()

        return self.annotate(
            unread_messages=
                Count('message__pk',
                filter=Q(
                    Q(message__room__member__user=user, message__created__gt=F('message__room__member__last_read')),
                    ~Q(message__user=user),
                    ~Q(message__user=None),
                )
            )
        )

    def empty(self):
        return self.model.objects.filter(message__isnull=True)

    def not_empty(self):
        return self.model.objects.exclude(message__isnull=True)

    def with_unread_messages(self, user):
        if not user.is_authenticated:
            return self.none()

        return self.of_user(user).count_unread_messages(user).filter(unread_messages__gte=1)

    def recent(self, user):
        if not user.is_authenticated:
            return self.none()

        return self.of_user(user).count_unread_messages(user).filter(
            Q(unread_messages__gte=1) |
            Q(modified__gte=now() - timedelta(days=settings.RECENT_ROOM_THRESHOLD))
        )

    def unread_and_not_notified(self, user):
        """
        - unread: [member.last_read < last message created]       # notify only about unread messages
        - unsent: [member.last_notified < room modified]          # do not notify about same messages
        - dont spam: [member.last_notified < now() - threshold]   # do not spam
        - ready to send: [member.last_read < now() - threshold]   # give user enough time to read message in app
        - user inactive (user.last_seen < now() - threshold]      # send notification if user is not active

        :return: Unread messages which haven't been sent by email yet in defined time threshold
        """
        if not user.is_authenticated:
            return self.none()

        threshold_date = now() - timedelta(minutes=settings.NOTIFY_ROOM_THRESHOLD)

        query = Q(
                member__user=user,
                member__last_notified__lt=threshold_date
            ) & \
            Q(
                member__user=user,
                member__last_read__lt=threshold_date
            ) & \
            Q(
                member__user=user,
                member__last_notified__lt=F('modified')
            )

        user_activity_attribute = settings.USER_ACTIVITY_ATTRIBUTE

        if user_activity_attribute:
            query &= Q(**{f'member__user__{user_activity_attribute}__lt': threshold_date})

        return self.with_unread_messages(user).filter(query)


class MessageQuerySet(QuerySet):
    def of_room(self, room):
        return self.filter(room=room)

    def unread_by_user(self, user):
        if not user.is_authenticated:
            return self.none()

        return self \
            .exclude(Q(user=user) | Q(user=None)) \
            .filter(
                room__member__user=user,
                created__gt=F('room__member__last_read')
            )



from django.contrib import admin
from django.db import models
from django.db.models import Count, Value
from django.db.models.functions import Substr, StrIndex
from django.utils.translation import ugettext_lazy as _
from whisper.models import Room, Message


class MessageInline(admin.StackedInline):
    model = Message
    extra = 1


class MemberInline(admin.TabularInline):
    model = Room.users.through
    verbose_name = _("Member")
    verbose_name_plural = _("Members")
    extra = 1


class EmptyRoomFilter(admin.SimpleListFilter):
    title = _('empty')
    parameter_name = 'empty'

    def lookups(self, request, model_admin):
        return (
            ('no', _('no')),
            ('yes', _('yes')),
        )

    def queryset(self, request, queryset):
        if self.value() == 'no':
            return queryset.not_empty()
        if self.value() == 'yes':
            return queryset.empty()


class SlugPrefixRoomFilter(admin.SimpleListFilter):
    title = _('slug prefix')
    parameter_name = 'slug_prefix'

    def lookups(self, request, model_admin):
        slug_prefixes = Room.objects.all()\
            .annotate(slug_prefix=Substr('slug', 1, StrIndex('slug',Value('-')) - 1, output_field=models.CharField()))\
            .order_by('slug_prefix')\
            .values_list('slug_prefix', flat=True).distinct()

        lookups = {}

        for prefix in slug_prefixes:
            lookups.update({prefix: _(prefix)})

        return lookups.items()

    def queryset(self, request, queryset):
        if self.value() is not None:
            return queryset.filter(slug__startswith=self.value())


class UsersNumberRoomFilter(admin.SimpleListFilter):
    title = _('number of users')
    parameter_name = 'number_of_users'

    def lookups(self, request, model_admin):
        return (
            ('1', '1 {}'.format(_('user'))),
            ('2', '2 {}'.format(_('users'))),
            ('3', '3 {}'.format(_('and more'))),
        )

    def queryset(self, request, queryset):
        try:
            value = int(self.value())
            lookup = 'gte' if value == 3 else 'exact'
            kwargs = {'count_users__{}'.format(lookup): value}
            qs = queryset.annotate(count_users=Count('users'))
            return qs.filter(**kwargs)
        except (ValueError, TypeError):
            return queryset


@admin.register(Room)
class RoomAdmin(admin.ModelAdmin):
    date_hierarchy = 'created'
    search_fields = ['name', 'slug']
    list_display = ['id', 'name', 'slug', 'created', 'modified']
    inlines = [MemberInline, MessageInline]
    list_filter = [EmptyRoomFilter, SlugPrefixRoomFilter, UsersNumberRoomFilter]



import json

from asgiref.sync import async_to_sync
from django import forms
from django.contrib.auth import get_user_model
from django.forms import ModelMultipleChoiceField
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from whisper.helpers import ChatMessageHelper
from whisper.models import Room, RoomUser


class RoomAddMemberForm(forms.ModelForm):
    class Meta:
        model = Room
        fields = ['users']

    def __init__(self, current_user=None, *args, **kwargs):
        room_pk = kwargs.pop('room_pk', None)
        super().__init__(*args, **kwargs)
        self.current_user = current_user

        if room_pk is not None:
            self.possible_users = RoomUser.get_possible_users_to_add(room_pk)
        else:
            self.possible_users = get_user_model().objects.all()

        self.fields['users'] = ModelMultipleChoiceField(required=True, queryset=self.possible_users)

    def clean(self):
        cleaned_data = super().clean()
        users = cleaned_data.get('users', None)
        users = set(users.values_list('pk', flat=True))
        current_users = RoomUser.get_users(self.instance)
        current_users = set(current_users.values_list('id', flat=True))

        if self.instance.is_user_to_user_room and not current_users.issubset(users):
            self.add_error('users', _('Cannot remove users from user to user chat'))

        if self.current_user.pk not in users:
            self.add_error('users', _('Cannot remove yourself from chat'))

        return cleaned_data

    def save(self, commit=True):
        users = self.cleaned_data.pop('users', None)
        current_users = RoomUser.get_users(self.instance)

        if self.instance.is_user_to_user_room:
            users = list(users.values_list('pk', flat=True))
            return Room.objects.get_or_create_from_users(users)
        else:
            for current_user in current_users:
                if current_user not in users:
                    async_to_sync(ChatMessageHelper.send_message)(self.instance, json.dumps({'USER_LEFT': {'username': str(current_user)}}))
                    RoomUser.objects.filter(user=current_user, room=self.instance).delete()

            if users is not None:
                for user in users:
                    if user not in current_users:
                        RoomUser.objects.update_or_create(room=self.instance, user=user, defaults={'last_read': now()})
                        async_to_sync(ChatMessageHelper.send_message)(self.instance, json.dumps({'USER_JOINED': {'username': str(user)}}))

            return super().save(commit)



import json
from json import JSONDecodeError
from channels.db import database_sync_to_async
from channels.layers import get_channel_layer
from django.template.defaultfilters import date
from django.templatetags.tz import localtime
from whisper import settings
from whisper.models import Message


class ChatMessageHelper:
    # send system message
    @staticmethod
    async def send_message(room, text_message, sender=None):
        message = Message.objects.create(
            room=room,
            user=sender,
            text=text_message,
        )

        # update modified timestamp
        room.save()

        channel_layer = get_channel_layer()

        # Send message to room group
        await channel_layer.group_send(
            room.group_name, {
                'type': 'chat_message',
                'message': str(message),
                'username': str(sender) if sender is not None else None,
                'timestamp': date(localtime(message.created), settings.DATETIME_FORMAT)
            }
        )

        for group_name in room.user_groups:
            await channel_layer.group_send(
                group_name, {
                    'type': 'chat_message',
                }
            )

        await ChatMessageHelper.send_room_properties(room, channel_layer)

    @staticmethod
    def message_from_type(type, **params):
        message = settings.MESSAGE_TYPES.get(type, type)

        try:
            return message.format(**params) if params is not None else message
        except AttributeError:
            return message

    @staticmethod
    async def send_room_properties(room, channel_layer):
        # send room properties
        user_count = await ChatMessageHelper.get_user_count(room)
        await channel_layer.group_send(
            room.group_name, {
                'type': 'room_properties',
                'room_name': room.name,
                'room_id': room.pk,
                'user_count': user_count
            })

    @staticmethod
    @database_sync_to_async
    def get_user_count(room):
        return room.users.count()



import django_rq
from django.apps import AppConfig
from django.conf import settings
from django.utils.translation import ugettext_lazy as _

from whisper.cron import notify_about_unread_messages


class ChatConfig(AppConfig):
    name = 'whisper'
    verbose_name = _('Chat')

    def ready(self):
        self.schedule_jobs()

    def schedule_jobs(self):
        print('Scheduling chat jobs...')

        try:
            import django_rq
            scheduler = django_rq.get_scheduler('cron')

            # Cron task to check status of companies
            scheduler.cron(
                "*/5 * * * *",  # Run every 5 minutes
                func=notify_about_unread_messages,
                timeout=settings.RQ_QUEUES['cron']['DEFAULT_TIMEOUT']
            )
        except ImportError:
            # django RQ couldn't be imported, the email notifications won't work
            pass
        except KeyError:
            # 'cron' not found in RQ queues
            # TODO: create whistle setting for this
            pass



import json
from json import JSONDecodeError

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.urls import reverse
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from whisper.managers import RoomQuerySet, MessageQuerySet


class Room(models.Model):
    slug = models.SlugField(unique=True, max_length=50)
    name = models.CharField(_('name'), max_length=200)
    users = models.ManyToManyField(to=settings.AUTH_USER_MODEL, verbose_name=_('users'), through='RoomUser')
    created = models.DateTimeField(_('created'), auto_now_add=True)
    modified = models.DateTimeField(_('modified'), auto_now=True)
    objects = RoomQuerySet.as_manager()

    class Meta:
        verbose_name = _('room')
        verbose_name_plural = _('rooms')
        ordering = ('-modified',)
        get_latest_by = 'created'

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        return reverse('whisper:room_detail', args=(self.slug,))

    @property
    def group_name(self):
        return f'room_{self.id}'

    @property
    def user_groups(self):
        user_groups = []

        for member in self.member_set.all():
            user_groups.append(f'unread-chat-messages-{member.user.pk}')

        return user_groups

    @property
    def is_user_to_user_room(self):
        return self.slug.startswith('users-')

    def add_user(self, user):
        return RoomUser.objects.get_or_create(room=self, user=user)[0]


class RoomUser(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    room = models.ForeignKey(Room, on_delete=models.CASCADE, related_query_name='member', related_name='member_set')
    last_read = models.DateTimeField(_('last read'), blank=True, null=True, default=None)
    last_notified = models.DateTimeField(_('last notified'), default=now)

    @staticmethod
    def get_possible_users_to_add(room_pk):
        return get_user_model().objects.exclude(roomuser__room__pk=room_pk)

    @staticmethod
    def get_users(room):
        return get_user_model().objects.filter(roomuser__room=room)

    @staticmethod
    def get_user(room, user_id):
        try:
            return get_user_model().objects.get(roomuser__room=room, roomuser__user__pk=user_id)
        except ObjectDoesNotExist:
            return None


class Message(models.Model):
    room = models.ForeignKey(Room, on_delete=models.CASCADE)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, blank=True, null=True, default=None)
    text = models.TextField(_('text'))
    created = models.DateTimeField(_('created'), auto_now_add=True)
    modified = models.DateTimeField(_('modified'), auto_now=True)
    objects = MessageQuerySet.as_manager()

    class Meta:
        verbose_name = _('message')
        verbose_name_plural = _('messages')
        ordering = ('created',)

    def __str__(self):
        if self.user is not None:
            # non-system message
            return self.text

        # system message
        from whisper.helpers import ChatMessageHelper
        try:
            message = json.loads(self.text)
            message_key = next(iter(message))
            params = message.get(message_key)
            return ChatMessageHelper.message_from_type(message_key, **params)
        except JSONDecodeError:
            return str(self.text)

    def get_absolute_url(self):
        return self.room.get_absolute_url()



# Generated by Django 2.0.6 on 2018-09-12 07:46

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('text', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': 'message',
                'verbose_name_plural': 'messages',
                'ordering': ('created',),
            },
        ),
        migrations.CreateModel(
            name='Room',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('slug', models.SlugField(unique=True)),
                ('name', models.CharField(max_length=200)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': 'room',
                'verbose_name_plural': 'rooms',
                'ordering': ('-modified',),
                'get_latest_by': 'created',
            },
        ),
        migrations.CreateModel(
            name='RoomUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('last_read', models.DateTimeField(blank=True, default=None, null=True)),
                ('last_notified', models.DateTimeField(default=django.utils.timezone.now)),
                ('room', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='member_set', related_query_name='member', to='whisper.Room')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='room',
            name='users',
            field=models.ManyToManyField(through='whisper.RoomUser', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='message',
            name='room',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='whisper.Room'),
        ),
        migrations.AddField(
            model_name='message',
            name='user',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model

from django.utils.translation import ugettext_lazy, ugettext

ROOM_FORM_CLASS = getattr(
    settings, 'WHISPER_ROOM_FORM_CLASS', None
)

ROOM_ADD_MEMBER_FORM_CLASS = getattr(
    settings, 'WHISPER_ROOM_ADD_MEMBER_FORM_CLASS', 'whisper.forms.RoomAddMemberForm'
)

USER_ACTIVITY_ATTRIBUTE = getattr(
    settings, 'WHISPER_USER_ACTIVITY_ATTRIBUTE', None
)

USER_INFO = getattr(
    settings, 'WHISPER_USER_INFO', lambda user: None if USER_ACTIVITY_ATTRIBUTE is None else getattr(user, USER_ACTIVITY_ATTRIBUTE)
)

CHAT_USERS = getattr(
    settings, 'WHISPER_CHAT_USERS', lambda: get_user_model().all()
)

RECENT_ROOM_THRESHOLD = getattr(
    settings, 'WHISPER_RECENT_ROOM_THRESHOLD', 7  # days
)

NOTIFY_ROOM_THRESHOLD = getattr(
    settings, 'WHISPER_NOTIFY_ROOM_THRESHOLD', 30  # minutes
)

DATETIME_FORMAT = getattr(
    settings, 'WHISPER_DATETIME_FORMAT', 'd.m.Y H:i:s'
)

MESSAGE_TYPES = getattr(
    settings, 'WHISPER_MESSAGE_TYPES',
    {
        'USER_LEFT': ugettext_lazy('{username} left room'),
        'USER_JOINED': ugettext_lazy('{username} joined room'),
        'USER_TYPING': ugettext('{username} is typing ...'),
    }
)



from whisper import settings
from whisper.models import Room


def chat(request):
    # TODO: cache to save DB requests or use template tags in chat.html template instead

    return {
        'chat_users': settings.CHAT_USERS,
        'recent_rooms': Room.objects.recent(request.user).not_empty()
    }



from django.core.exceptions import ObjectDoesNotExist
from django.utils.functional import cached_property

from whisper.models import Room


class ChatRoomMixin(object):
    @cached_property
    def room_members(self):
        raise NotImplementedError()

    @property
    def room_slug(self):
        raise NotImplementedError()

    @property
    def room(self):
        try:
            return Room.objects.get(slug=self.room_slug)
        except ObjectDoesNotExist:
            return None



from django.urls import path
from django.utils.translation import pgettext_lazy

from whisper.views import RoomView, RoomUpdateView, RoomLeaveView, RoomAddMemberView, RoomListView

app_name = 'whisper'

urlpatterns = [
    path(pgettext_lazy('url', 'room/<slug>/update/'), RoomUpdateView.as_view(), name='room_update'),
    path(pgettext_lazy('url', 'room/<slug>/leave/'), RoomLeaveView.as_view(), name='room_leave'),
    path(pgettext_lazy('url', 'room/<slug>/add-member/'), RoomAddMemberView.as_view(), name='room_add_member'),
    path(pgettext_lazy('url', 'room/<slug>/'), RoomView.as_view(), name='room_detail'),
    path(pgettext_lazy('url', 'room/'), RoomListView.as_view(), name='room_list'),
]



from datetime import timedelta

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.template import loader, TemplateDoesNotExist
from django.utils import translation
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _


def notify_about_unread_messages():
    """
    - iterate every user and check their rooms with unread and not notified messages
    """
    from whisper.models import Room, RoomUser

    notified_users = 0

    # TODO: not logged/away in only?
    for user in get_user_model().objects.all():
        unread_and_not_notified_rooms = Room.objects.unread_and_not_notified(user)

        if not unread_and_not_notified_rooms.exists():
            continue

        notify(user, unread_and_not_notified_rooms)
        notified_users += 1

        for room in unread_and_not_notified_rooms:
            member = RoomUser.objects.get(room=room, user=user)
            member.last_notified = now()
            member.save(update_fields=['last_notified'])

    return notified_users


def notify(user, rooms):
    """
    Send email notification about unread messages to room member
    """

    # activate language
    language = settings.LANGUAGE_CODE  # TODO: pick from user profile
    translation.activate(language)

    # add unread messages to room object
    for room in rooms:
        room.unread_messages = room.message_set.unread_by_user(user)

    # templates
    t = loader.get_template('whisper/mails/notification.txt')

    try:
        t_html = loader.get_template('whisper/mails/notification.html')
    except TemplateDoesNotExist:
        t_html = None

    # recipients
    recipient = user
    recipient_list = [recipient.email]

    # description
    description = _('You have unread chat messages')

    # subject
    site = get_current_site(None)

    subject = '[{}] {}'.format(
        site.name,
        description
    )

    # context
    context = {
        'subject': subject,
        'description': description,
        'recipient': recipient,
        'rooms': rooms,
    }

    # message
    message = t.render(context)
    html_message = t_html.render(context) if t_html else None

    # mail arguments
    from_email = settings.DEFAULT_FROM_EMAIL
    fail_silently = True

    send_mail(subject, message, from_email, recipient_list, html_message=html_message, fail_silently=fail_silently)



from django.conf.urls import url

from . import consumers


websocket_urlpatterns = [
    url(r'^ws/chat/unread-messages/$', consumers.UnreadChatMessagesConsumer),
    url(r'^ws/chat/(?P<room_slug>[^/]+)/$', consumers.ChatConsumer),
]



from whisper import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect
from django.utils.module_loading import import_string
from django.views.generic import DetailView, UpdateView, ListView
from whisper.models import Room, RoomUser


class RoomView(LoginRequiredMixin, DetailView):
    model = Room


class RoomListView(LoginRequiredMixin, ListView):
    model = Room

    def get_queryset(self):
        return super().get_queryset().recent(self.request.user)


class RoomLeaveView(LoginRequiredMixin, DetailView):
    model = Room

    def dispatch(self, request, *args, **kwargs):
        room = self.get_object()
        user = request.user
        RoomUser.objects.filter(user=user, room=room).delete()
        return redirect('/')


class RoomUpdateView(LoginRequiredMixin, UpdateView):
    model = Room
    fields = ['name']

    def get_form_class(self):
        form_class = settings.ROOM_FORM_CLASS
        return import_string(form_class) if form_class else super().get_form_class()


class RoomAddMemberView(LoginRequiredMixin, UpdateView):
    model = Room
    fields = ['users']
    template_name = 'whisper/add_member_form.html'

    def dispatch(self, request, *args, **kwargs):
        self.user = request.user
        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update({'current_user': self.user})
        return kwargs

    @staticmethod
    def load_form_class():
        form_class = settings.ROOM_ADD_MEMBER_FORM_CLASS

        if form_class:
            form_class = import_string(form_class)

        return form_class

    def get_form_class(self):
        return self.load_form_class()



import json

from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.template import loader
from django.template.defaultfilters import date
from django.templatetags.tz import localtime
from django.utils.timezone import now

from whisper import settings
from whisper.helpers import ChatMessageHelper
from whisper.models import Room, Message, RoomUser
from whisper.views import RoomAddMemberView


class ChatConsumer(AsyncJsonWebsocketConsumer):

    async def websocket_connect(self, message):
        room_slug = self.scope['url_route']['kwargs']['room_slug']

        # init variables
        self.user = self.scope['user']
        self.room = await self.get_room(room_slug)
        self.groups.append(self.room.group_name)
        type = message['type']

        # update last read flag of current room user
        if await self.get_room_user() is None:
            await ChatMessageHelper.send_message(self.room, json.dumps({'USER_JOINED': {'username': str(self.user)}}))

        await self.update_room_user(self.room, self.user)
        await super().websocket_connect(message)

        # send room properties
        self.user_count = await ChatMessageHelper.get_user_count(self.room)
        await self.send_json(content={
            'type': 'room_properties',
            'room_name': self.room.name,
            'room_id': self.room.pk,
            'room_slug': room_slug,
            'room_modified': date(localtime(self.room.modified), settings.DATETIME_FORMAT),
            'user_count': self.user_count,
            'is_user_to_user': self.room.is_user_to_user_room
        })

        # init room with previous messages
        for message in await self.get_room_messages():
            await self.send_json(content={
                'type': type,
                'message': str(message),
                'timestamp': date(localtime(message.created), settings.DATETIME_FORMAT),
                'username': str(message.user) if message.user is not None else None,
            })

        await self.channel_layer.group_send(
            f'unread-chat-messages-{self.user.pk}', {
                'type': 'chat_message',
            }
        )

    @database_sync_to_async
    def get_room(self, slug):
        try:
            return Room.objects.get(slug=slug)
        except ObjectDoesNotExist:
            return Room.objects.create_from_slug(slug, self.user)

    @database_sync_to_async
    def get_room_messages(self):
        return self.room.message_set.all()

    @database_sync_to_async
    def remove_user_from_room(self, user_id):
        return RoomUser.objects.filter(user__pk=user_id, room=self.room).delete()

    @database_sync_to_async
    def get_user(self, user_id):
        return RoomUser.get_user(self.room, user_id)

    @database_sync_to_async
    def get_room_users(self):
        return RoomUser.get_users(self.room)

    @database_sync_to_async
    def add_room_users(self, user_pks):
        users = []

        user_objects = get_user_model().objects.filter(pk__in=user_pks)

        for user in user_objects:
            users.append(self.room.add_user(user))

        return users

    @database_sync_to_async
    def get_room_user(self):
        try:
            return RoomUser.objects.get(room=self.room, user=self.user)
        except ObjectDoesNotExist:
            return None

    @database_sync_to_async
    def update_room_user(self, room, user):
        RoomUser.objects.update_or_create(room=room, user=user, defaults={'last_read': now()})

    @database_sync_to_async
    def get_or_create_group_room(self, user_ids):
        user_ids.extend(list(self.room.users.all().values_list('pk', flat=True)))
        return Room.objects.get_or_create_from_users(user_ids)

    # Receive message from WebSocket
    async def receive_json(self, content):
        if self.user.is_authenticated:
            json_type = content.get('type', None)

            if json_type == 'leave_room':
                await self.remove_user_from_room(self.user.pk)
                dict_message = {'USER_LEFT': {'username': str(self.user), 'room': self.room.name, "timestamp": date(localtime(now()), settings.DATETIME_FORMAT)}}
                await ChatMessageHelper.send_message(self.room, json.dumps(dict_message))

            elif json_type == 'user_typing':

                for group_name in self.groups:
                    await self.channel_layer.group_send(
                        group_name, {
                            'type': 'user_typing',
                            'username': str(self.user),
                            'text': ChatMessageHelper.message_from_type('USER_TYPING', username=str(self.user))
                        }
                    )

            elif json_type == 'room_members':
                await self.send_room_members()

            elif json_type == 'remove_member':
                user_ids = content.get('user_id', None)
                user = await self.get_user(user_ids)
                await self.remove_user_from_room(user_ids)
                dict_message = {'USER_LEFT': {'username': str(user), 'room': self.room.name, "timestamp": date(localtime(now()), settings.DATETIME_FORMAT)}}
                await ChatMessageHelper.send_message(self.room, json.dumps(dict_message))

            elif json_type == 'add_members':
                user_ids = content.get('user_ids', None)

                if user_ids is not None:
                    user_ids = list(map(int, user_ids))

                    if self.room.is_user_to_user_room:
                        new_room = await self.get_or_create_group_room(user_ids)

                        for group_name in self.groups:
                            await self.channel_layer.group_send(
                                group_name, {
                                    'type': 'new_room',
                                    'slug': new_room.slug,
                                }
                            )
                    else:
                        users = await self.add_room_users(user_ids)

                        for user in users:
                            dict_message = {'USER_JOINED': {'username': str(user.user)}}
                            await ChatMessageHelper.send_message(self.room, json.dumps(dict_message))

                        await self.send_room_members()

            else:
                text = content['message']

                # update last read flag of current room user (create if not exists)
                await self.update_room_user(self.room, self.user)

                message = Message.objects.create(
                    room=self.room,
                    user=self.user,
                    text=text
                )

                # update modified timestamp
                self.room.save()

                # Send message to room group
                for group_name in self.groups:
                    await self.channel_layer.group_send(
                        group_name, {
                            'type': 'chat_message',
                            'message': text,
                            'username': str(self.user),
                            'timestamp': date(localtime(message.created), settings.DATETIME_FORMAT)
                        }
                    )

                for group_name in self.room.user_groups:
                    await self.channel_layer.group_send(
                        group_name, {
                            'type': 'chat_message',
                        }
                    )

    async def send_room_members(self):
        users = await self.get_room_users()
        form_class = RoomAddMemberView.load_form_class()
        await self.send_json(content={
            'type': 'room_members',
            'form': str(form_class(room_pk=self.room.pk).as_p()),
            'members': [{
                'id': user.id,
                'name': str(user),
                'html': loader.get_template('whisper/member.html').render(
                    {'user': user, 'scope_user': self.user, 'is_user_to_user': self.room.is_user_to_user_room}
                )
            } for user in users],
        })

    # Receive message from room group
    async def chat_message(self, event):
        # Send message to WebSocket
        await self.send_json(content=event)

    # Receive user_typing from room group
    async def user_typing(self, event):
        # Send user_typing to WebSocket
        await self.send_json(content=event)

    # Receive room_properties from room group
    async def room_properties(self, event):
        # Send room_properties to WebSocket
        await self.send_json(content=event)

    # Receive new_room from room group
    async def new_room(self, event):
        # Send new_room to WebSocket
        await self.send_json(content=event)


class UnreadChatMessagesConsumer(AsyncJsonWebsocketConsumer):
    async def websocket_connect(self, message):
        self.user = self.scope['user']

        if self.user.is_authenticated:
            self.group_name = f'unread-chat-messages-{self.user.pk}'
            self.groups.append(self.group_name)

            await super().websocket_connect(message)

            unread_messages = await self.get_unread_messages(self.scope['user'])

            # init current socket
            await self.send_json(content=unread_messages)

    @database_sync_to_async
    def get_unread_messages(self, user):
        unread_messages = Message.objects.unread_by_user(user).count()
        unread_rooms = Room.objects.with_unread_messages(user).values('pk', 'unread_messages')

        return {
            'unread_messages': unread_messages,
            'unread_rooms': list(unread_rooms)
        }

    # Receive message from room group
    async def chat_message(self, event):
        # Send message to WebSocket
        unread_messages = await self.get_unread_messages(self.scope['user'])
        await self.send_json(content=unread_messages)



#!/usr/bin/env python
from setuptools import setup


setup(
    name='django-whisper',
    version='0.2.0',
    description='Chat with rooms for Django based on websockets',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Pragmatic Mates',
    author_email='info@pragmaticmates.com',
    maintainer='Pragmatic Mates',
    maintainer_email='info@pragmaticmates.com',
    url='https://github.com/PragmaticMates/django-whisper',
    packages=[
        'whisper',
        'whisper.migrations'
    ],
    include_package_data=True,
    install_requires=('django', 'channels'),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'License :: OSI Approved :: BSD License',
        'Development Status :: 3 - Alpha'
    ],
    license='BSD License',
    keywords="django chat room channels websocket",
)
